/**********************************************************************
 * AVMA.C       Copyright (C) University of Lincs and Humberside 1999 *
 *--------------------------------------------------------------------*
 * AVM - Assembler for a Virtual Machine                              *
 *--------------------------------------------------------------------*
 * Version: 1.00;  Author: Damian G Walker;  Date: 20-Feb-1999        *
 **********************************************************************/

#define _AVMASM_C_

/* included headers ***************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/* defined constants **************************************************/

#define SYMSIZE 128

/* enumerated types ***************************************************/

typedef enum {			/* addressing modes */
  ASM_IMMEDIATE,		/* immediate addressing */
  ASM_IMPLIED,			/* implied addressing */
  ASM_INDIRECT,			/* indirect addressing */
  ASM_DIRECT = ASM_INDIRECT,	/* like indirect as far as ASM is concerned */
  ASM_RELATIVE,			/* relative addressing */
  ASM_SPECIAL			/* special mnemonics */
} avma_mode;

/* data structures ****************************************************/

typedef struct {		/* symbol table */
  char name[7];			/* name of symbol */
  int value;			/* value of symbol */
} avma_symbol;

typedef struct {		/* AVM instruction */
  char *mnemonic;		/* mnemonic for the instruction */
  int mode,			/* addressing mode */
  opcode;			/* opcode for instruction */
} avma_instruction;

/* global variables ***************************************************/

FILE *asmfile, /* assembly language file */
*imgfile; /* image file */
unsigned int address; /* current address */
avma_symbol symbol[SYMSIZE]; /* symbol table */
avma_instruction inst[] = {	/* instruction list */
  { "ADD", ASM_IMPLIED, 0x78 },
  { "ADD", ASM_IMMEDIATE, 0x08 },
  { "ADD", ASM_INDIRECT, 0x80 },
  { "AND", ASM_IMPLIED, 0x79 },
  { "AND", ASM_IMMEDIATE, 0x09 },
  { "AND", ASM_INDIRECT, 0x90 },
  { "CMP", ASM_IMPLIED, 0x7e },
  { "CMP", ASM_IMMEDIATE, 0x0e },
  { "CMP", ASM_INDIRECT, 0xe0 },
  { "CSR", ASM_RELATIVE, 0x06 },
  { "CSR", ASM_DIRECT, 0x60 },
  { "DEC", ASM_IMPLIED, 0x7a },
  { "INC", ASM_IMPLIED, 0x72 },
  { "JCA", ASM_RELATIVE, 0x03 },
  { "JEQ", ASM_RELATIVE, 0x01 },
  { "JMP", ASM_RELATIVE, 0x02 },
  { "JMP", ASM_DIRECT, 0x20 },
  { "JNC", ASM_RELATIVE, 0x07 },
  { "JNE", ASM_RELATIVE, 0x05 },
  { "LDA", ASM_IMMEDIATE, 0x40 },
  { "LDA", ASM_INDIRECT, 0x10 },
  { "LDD", ASM_IMPLIED, 0x7b },
  { "LDD", ASM_IMMEDIATE, 0x0b },
  { "LDD", ASM_INDIRECT, 0xb0 },
  { "ORD", ASM_IMPLIED, 0x7d },
  { "ORD", ASM_IMMEDIATE, 0x0d },
  { "ORD", ASM_INDIRECT, 0xd0 },
  { "PHA", ASM_IMPLIED, 0x70 },
  { "PHD", ASM_IMPLIED, 0x71 },
  { "PPA", ASM_IMPLIED, 0x74 },
  { "PPD", ASM_IMPLIED, 0x75 },
  { "RET", ASM_IMPLIED, 0x76 },
  { "STA", ASM_DIRECT, 0x50 },
  { "STD", ASM_IMPLIED, 0x73 },
  { "STD", ASM_DIRECT, 0x30 },
  { "SUB", ASM_IMPLIED, 0x7c },
  { "SUB", ASM_IMMEDIATE, 0x0c },
  { "SUB", ASM_INDIRECT, 0xc0 },
  { "XOR", ASM_IMPLIED, 0x7f },
  { "XOR", ASM_IMMEDIATE, 0x0f },
  { "XOR", ASM_INDIRECT, 0xf0 },
  { "ORG", ASM_SPECIAL, 0x00 }, /* origin of code */
  { "DEF", ASM_SPECIAL, 0x01 }, /* definition of bytes */
  { "EQU", ASM_SPECIAL, 0x02 }, /* macro assignment */
  { "", 0, 0 }			/* place marker */
};

/* level 5 routines ***************************************************/

/* look up a value from the symbol table */
static int avma_symlookup(int *retval, char *value) {
  int symcount;			/* used to look through symbol list */
  for(symcount = 0; symcount < SYMSIZE; ++symcount)
    if( !strcmp(value, symbol[symcount].name) )
      break;
  if(symcount < SYMSIZE) {
    *retval = symbol[symcount].value;
    return 1;
  }
  return 0;
}

/* add a new symbol to the symbol table */
static int avma_symadd(char *value) {
  int symcount;			/* used to look through symbol list */
  for(symcount = 0; symcount < SYMSIZE; ++symcount)
    if( !*symbol[symcount].name )
      break;
  if( !*symbol[symcount].name ) {
    strcpy( symbol[symcount].name, value );
    symbol[symcount].value = 0;
  }
  return 0;
}

/* level 4 routines ***************************************************/

/* evaluate a normal data value */
static int avma_val(char *value) {
  int retval;
  if( sscanf(value, "%d", &retval) == 1 )
    return retval;
  else if( sscanf(value, "$%x", &retval) == 1 )
    return retval;
  else if( avma_symlookup(&retval, value) )
    return retval;
  else
    return avma_symadd(value);
}

/* evaluate a relative address */
static int avma_rval(char *value) {
  int retval;
  if( sscanf(value, "+%d", &retval) == 1 )
    return retval;
  else if( sscanf(value, "-%d", &retval) == 1 )
    return 0x100 - retval;
  else if( sscanf(value, "+$%x", &retval) == 1 )
    return retval;
  else if( sscanf(value, "-$%x", &retval) == 1 )
    return 0x100 - retval;
  else if( avma_symlookup(&retval, &value[1]) )
    return (retval - (address + 2)) & 0xff;
  else
    return avma_symadd(value);
}

/* process an EQU instruction */
static void avma_equ(char *label, int operand) {
  int symcount;			/* used to look through symbol list */
  for(symcount = 0; symcount < SYMSIZE; ++symcount)
    if( !strcmp(label, symbol[symcount].name) || !*symbol[symcount].name )
      break;
  if(symcount < SYMSIZE) {
    strcpy(symbol[symcount].name, label);
    symbol[symcount].value = operand;
  }
}

/* level 3 routines ***************************************************/

/* read a line from the assembly language file */
static int avma_getline(char *buf) {
  if( fgets(buf, 128, asmfile) )
    return 0;
  else
    return 1;
}

/* split a line into label, mnemonic and data */
static int avma_splitline(char *label, char *mnemonic, char *argument,
			   char *buf) {
  char part[3][128],		/* 3 parts of the line */
  *ptr;				/* temporary pointer */
  int items;			/* number of items found on the line */
  items = sscanf(buf, "%s%s%s", part[0], part[1], part[2]);
  switch(items) {
  case 0: case -1: *label = *mnemonic = *argument = 0; break;
  case 1:
    *label = *argument = 0;
    strcpy(mnemonic, part[0]);
    break;
  case 2:
    if(( ptr = strchr(part[0], ':') )) {
      *ptr = 0;
      strcpy(label, part[0]);
      strcpy(mnemonic, part[1]);
      *argument = 0;
    } else {
      *label = 0;
      strcpy(mnemonic, part[0]);
      strcpy(argument, part[1]);
    }
    break;
  case 3:
    strcpy(label, part[0]);
    if(( ptr = strchr(label, ':') )) *ptr = 0;
    strcpy(mnemonic, part[1]);
    strcpy(argument, part[2]);
    break;
  }
  return items;
}

/* process a label */
static void avma_processlabel(char *label) {
  int symcount;			/* symbol counter */
  for(symcount = 0; symcount < SYMSIZE; symcount++)
    if( !strcmp( symbol[symcount].name, label ) || !*symbol[symcount].name )
      break;
  if(symcount < SYMSIZE) {
    strcpy( symbol[symcount].name, label);
    symbol[symcount].value = address;
  }
}

/* process the argument */
static void avma_processarg(int *mode, int *operand, char *argument) {
  switch(*argument) {
  case 0: *mode = ASM_IMPLIED; *operand = 0; break;
  case '#': *mode = ASM_IMMEDIATE; *operand = avma_val(&argument[1]); break;
  case '+': case '-':
    *mode = ASM_RELATIVE; *operand = avma_rval(argument); break;
  case '@': *mode = ASM_SPECIAL; *operand = avma_val(&argument[1]); break;
  default:
    *mode = ASM_INDIRECT; *operand = avma_val(argument);
  }
}

/* process the mnemonic */
static int avma_processmnemonic(int pass, unsigned int *instruction, int mode, 
				 int operand, char *mnemonic, char *label) {
  int icount;			/* instruction counter */
  for(icount = 0; *inst[icount].mnemonic; ++icount)
    if( !strcmp(mnemonic, inst[icount].mnemonic) && inst[icount].mode == mode )
      break;
    else if( !strcmp(mnemonic, inst[icount].mnemonic) &&
	    inst[icount].mode == ASM_SPECIAL) {
      mode = ASM_SPECIAL;
      break;
    }
    else if( !strcmp(mnemonic, inst[icount].mnemonic) &&
	     mode == ASM_INDIRECT &&
	     inst[icount].mode == ASM_RELATIVE && 
	     (signed)(operand - (address + 2)) >= -128 &&
	       (signed)(operand - (address + 2)) <= 127) {
      mode = ASM_RELATIVE;
      operand = ( operand - (address + 2) ) & 0xff;
      break;
    }
    else if( !strcmp(mnemonic, inst[icount].mnemonic) &&
	     mode == ASM_INDIRECT &&
	     inst[icount].mode == ASM_RELATIVE && 
	     pass == 1) {
      mode = ASM_RELATIVE;
      operand = 0; /* relative address unimportant on 1st pass */
      break;
    }
  if( *inst[icount].mnemonic )
    switch(mode) {
    case ASM_SPECIAL:
      switch( inst[icount].opcode ) {
      case 0: address = operand; *instruction = 0; break;
      case 1: *instruction = operand; mode = ASM_IMPLIED; break;
      case 2: avma_equ(label, operand); break;
      }
      break;
    case ASM_IMPLIED: *instruction = inst[icount].opcode; break;
    default: *instruction = (inst[icount].opcode << 8) + operand; break;
    }
  else {
    mode = ASM_SPECIAL;
    *instruction = 0;
  }
  return mode;
}

/* write the mnemonic to the image file */
static void avma_write(unsigned int instruction, int mode) {
  if( ftell(imgfile) == 0 ) fputc(address >> 4, imgfile);
  switch(mode) {
  case ASM_SPECIAL: break;
  case ASM_IMPLIED: fputc(instruction, imgfile); break;
  default:
    fputc(instruction >> 8, imgfile);
    fputc(instruction & 0xff, imgfile);
  }
}

/* advance the address counter */
static void avma_step(int mode) {
  switch(mode) {
  case ASM_SPECIAL: break;
  case ASM_IMPLIED: ++address; break;
  default: address += 2;
  }
}

/* level 2 routines ***************************************************/

/* initialise a pass of the assembler */
static void avma_initprocess(void) {
  fseek(asmfile, 0, SEEK_SET);
  fseek(imgfile, 0, SEEK_SET);
  address = 0;
}

/* interpret a single line */
static int avma_processline(int pass) {
  char buf[128],		/* input buffer */
  *ptr,				/* pointer to semi-colon */
  label[128],			/* label */
  mnemonic[128],		/* mnemonic */
  argument[128];		/* argument */
  int mode,			/* apparent addressing mode */
  operand,			/* apparent operand */
  instruction;			/* opcode and operand byte or word */
  if( avma_getline(buf) ) return 1;
  if(( ptr = strchr(buf, ';') )) *ptr = 0;
  if( !avma_splitline(label, mnemonic, argument, buf) ) return 0;
  avma_processlabel(label);
  avma_processarg(&mode, &operand, argument);
  mode = avma_processmnemonic(pass, &instruction, mode, operand, mnemonic, 
			      label);
  if(pass == 2 && mode != ASM_SPECIAL)
    avma_write(instruction, mode);
  avma_step(mode);
  return 0;
}

/* clean up after a single pass */
static void avma_closeprocess(int pass) {
  int symcount; /* used to count through symbol table */
  if(pass != 2) return;
  for(symcount = 0; symcount < SYMSIZE; ++symcount)
    if( *symbol[symcount].name )
      printf("%s:\tEQU $%X\n", symbol[symcount].name, symbol[symcount].value);
}

/* level 1 routines ***************************************************/

/* initialise program */
static int avma_init(char *asmname, char *imgname) {
  int symcount;			/* used to count through symbol table */
  if(!( asmfile = fopen(asmname, "r") )) return 1;
  if(!( imgfile = fopen(imgname, "wb") )) {
    fclose(asmfile);
    return 1;
  }
  for(symcount = 0; symcount < SYMSIZE; symcount++)
    symbol[symcount].value = *symbol[symcount].name = 0;
  return 0;
}

/* perform a single assembler pass */
static void avma_process(int pass) {
  avma_initprocess();
  while( !avma_processline(pass) );
  avma_closeprocess(pass);
}

/* close the assembler down */
static void avma_close(void) {
  fclose(imgfile);
  fclose(asmfile);
}

/* level 0 routines ***************************************************/

/* main program */
int main(int argc, char **argv) {
  if(argc != 3) {
    puts("usage: avma ASMFILE IMGFILE");
    return 1;
  }
  if( avma_init(argv[1], argv[2]) ) return 1;
  avma_process(1);		/* pass 1 */
  avma_process(2);		/* pass 2 */
  avma_close();
  return 0;
}
